/*
 * Diese Datei enthält die Wireworld-spezifischen Ergänzungen zum allgemeinen
 * Zellularautomatencode in common.js, hauptsächlich die Regelanwendungen, sowie
 * den Parsercode für die Datei-laden/speichern-Funktionen.
 */

// Erweitere die Zellfarbenliste mit den Wireword-Zelltypen:
Settings.cellColor.push("#bb7733");  // "Kupferfarben" für Draht-Zellen.
Settings.cellColor.push("#77bbff");  // Hellblau für Elektronenenden.
Settings.cellColor.push("#0066ff");  // Dunkelblau für Elektronenköpfe.

// changeCellState() erweitert den Cell-Datentypen um eine Methode zum Ändern
// des Zellstatus. Die Funktion gibt in 'success' einen booleschen Wahrheitswert
// zurück: 'true', wenn der Status tatsächlich geändert wurde, 'false', wenn die
// Zelle sich bereits im gewünschten Zustand befindet. Außerdem wird in
// 'oldState' der alte Status der Zelle zurückgegeben, damit im aufrufenden Code
// ggf. die Populationsanzeige aktualisiert werden kann.
Cell.prototype.changeCellState = function (cellState)
{
    var oldState = this.state;
    var success  = false;

    if (oldState != cellState) {
        this.state = cellState;
        success = true;
    }

    return [success, oldState];
}


// WireField ist ein um Wireworld-spezifische Funktionen ergänzter Field-
// Datentyp, er "erbt" also von Field (in OOP-Terminologie):
var WireField = function (rows, columns)
{
    // Damit ein neues WireField-Objekt korrekt initialisiert wird, d.h. nach
    // dem Erzeugen wie ein Field-Objekt aussieht, muss im WireField-Konstruktor
    // (d.h. in dieser Funktion) die Konstruktor-Funktion des Field-Typen
    // aufgerufen werden. Damit dies mit der Referenz zum neuen WireField-Objekt
    // geschieht, muss dies über die Field.call()-Methode geschehen (die alle
    // neuen Objekte ihrerseits vom JavaScript-Standard-Objekttypen erben und
    // die genau für diesen Zweck vorhanden ist). Dort kann als erster Parameter
    // von Hand die neue Objektreferenz übergeben werden ('this' verweist dann
    // auf das hier erzeugte WireField-Objekt):
    Field.call(this, rows, columns);
}
// Mit der folgenden Zeile "erbt"/übernimmt der WireField-Prototyp den Inhalt
// des Field-Prototypen, d.h. alle WireField-Objekte basieren auf dem Prototypen
// des Field-Objekts (genau dies besagt "Vererbung" in der OOP-Terminologie):
WireField.prototype = Object.create(Field.prototype);
// Durch die vorige Zeile würde auch der oben definierte WireField-Konstruktor
// überschrieben (mit dem Konstruktor der Basis-Klasse). Mit der folgenden Zeile
// setzen wir ihn wieder auf den korrekten, oben definierten
// WireField-Konstruktor zurück.
WireField.prototype.constructor = WireField;

// WireField.createCell() verändert den Status einer Zelle im internen Cell-
// Array und ruft zu diesem Zweck ihrerseits die oben definierte
// Cell.changeCellState()-Funktion auf. Anhand des von dort erhaltenen
// Rückgabewerts wird ggf. die Populationsanzeige aktualisiert. Wenn der
// Funktion in 'update' ein Wert ungleich 0 übergeben wird, wird dieser (bei
// Erfolg der Zellstatusänderung) zur Aktualisierung der umgebenden Nachbarwerte
// verwendet. Durch diesen zusätzlichen Parameter ist die Funktion allgemein
// gehalten, dass sie sowohl während des interaktiven Zeichnens auf dem Canvas
// verwendet werden kann (hier wird der Funktion in 'update' ein entsprechender
// Wert übergeben), als auch während der Regelanwendung zur Berechnung der
// nächsten Generation. Hier wird 'update' bei der Statusänderung weggelassen
// und erst nach Änderung aller Zellzustände über countAllNeighbours() global
// neu gezählt. Auf diese Art wird nur ein einzelnes Spielfeld gebraucht und die
// Unterscheidung zwischen Spielfeld und Arbeitsfeld, zwischen denen hin- und
// herkopiert wird, entfällt.
WireField.prototype.createCell = function (row, column, cellState, update)
{
    // Rufe die zelleigene .changeCellState()-Methode auf.
    var retval   = this.data[row][column].changeCellState(cellState);
    // Teile den erhaltenen Rückgabewert-Array auf (zur besseren
    // Verständlichkeit des folgenden Codes):
    var success  = retval[0];
    var oldState = retval[1];

    // Wenn die Statusänderung erfolgreich war, d.h. die Zelle sich noch nicht
    // im gewünschten Status befand:
    if (success) {
        if (cellState == 3)
            // Wenn cellState == 3, dann haben wir einen neuen Elektronenkopf im
            // Universum und müssen die Population erhöhen:
            this.population++;
        else if (oldState == 3)
            // Wenn hingegen oldState == 3, dann hatte die Statusänderung zur
            // Folge, dass nun ein Elektronenkopf weniger existiert:
            this.population--;

        // Wir aktualisieren die umgebenden Nachbarn nur, wenn uns der
        // aufrufende Code darum bittet (dann haben wir im Parameter 'update'
        // einen Wert ungleich 0 erhalten).
        if (update) {
            if (cellState == 3) {
                this.updateNeighbours(row, column, update);
                this.updatePopulationHTML();
            }
            else if (oldState == 3) {
                this.updateNeighbours(row, column, -1);
                this.updatePopulationHTML();
            }
        }

        // Die Statusänderung war erfolgreich, deswegen geben wir 'true' zurück:
        return true;
    } else
        // Die Statusänderung schlug fehl, deswegen geben wir 'false' zurück:
        return false;
}

// WireField.countAllNeighbours() wird von .nextGeneration() benötigt, um nach
// der Regelanwendung die Nachbarwerte der neu berechneten Generation zu
// aktualisieren:
WireField.prototype.countAllNeighbours = function ()
{
    // Damit wir korrekt zählen, müssen wir die Nachbarwerte erst zurücksetzen:
    this.resetAllNeighbours();

    // Wir müssen den äußeren, unsichtbaren Zellrand weglassen, deswegen
    // beginnen wir bei 1:
    for (var row = 1; row <= this.rows; row++) {
        for (var column = 1; column <= this.columns; column++) {
            // Nur Cell.state == 3 hat Auswirkungen auf die Nachbarwerte:
            if (this.data[row][column].state == 3)
                this.updateNeighbours(row, column, 1);
        }
    }
}

// WireField.nextGeneration() ist die wohl wichtigste Funktion des ganzen
// Programms, denn hier findet die Wireworld-Regelanwendung statt.
WireField.prototype.nextGeneration = function ()
{
    // Wir müssen den äußeren, unsichtbaren Zellrand weglassen, deswegen
    // beginnen wir bei 1:
    for (var row = 1; row <= this.rows; row++) {
        for (var column = 1; column <= this.columns; column++) {
            // Zur besseren Lesbarkeit:
            var currentState = this.data[row][column].state;

            switch (currentState) {
            case 1:  // '1' ist der Drahtzelltyp
                // Wenn genau 1 oder 2 Nachbarzellen Elektronenköpfe sind, dann
                // wird diese Drahtzelle ebenfalls zu einem Elektronenkopf:
                if (this.data[row][column].neighbours == 1 ||
                    this.data[row][column].neighbours == 2) {

                    // this.createCell sorgt für die interne Zustandsverwaltung.
                    // Wir übergeben keinen Wert für den Parameter 'update', da
                    // sonst die Nachbarwerte für die nächsten Zellen nicht mehr
                    // stimmen würden:
                    this.createCell(row, column, 3);

                    // drawCell besorgt die visuelle Darstellung auf dem Canvas:
                    drawCell(row, column, 3);
                } // Andernfalls passiert nichts.
                break;
            case 2:  // '2' ist ein Elektronenende,
            case 3:  // '3' ist ein Elektronenkopf.
                // Der Regelcode für diese beiden Zelltypen ist identisch, da
                // sie jeweils nur für eine Generation in ihrem Status
                // bleiben. Durch die Wahl der Zustandswerte 2 und 3 für
                // Elektronenzellen (und 1 für die Drahtzellen) wird mit einem
                // einfachen 'currentState - 1' der korrekte Zustandswert für
                // die kommende Genreation ermittelt:
                this.createCell(row, column, currentState - 1);
                drawCell(row, column, currentState - 1);
            case 0:  // 'leere' Zellen bleiben in Wireworld leer.
            default:
                break;
            }
        }
    }

    // Wir haben alle neuen Zustände berechnet, jetzt müssen die Nachbarwerte
    // aktualisiert werden...
    this.countAllNeighbours();

    // ...ebenso die Generations sowie auch die Populationsanzeige:
    this.increaseGeneration();
    this.updatePopulationHTML();
}


// Das eigentliche Spielfeld wird hier als globale Variable erzeugt, damit es
// von allen Stellen im Code erreicht werden kann:
var field = new WireField(Settings.cellsPerLine, Settings.cellsPerLine);


/*
 * Ende des Wireworld-spezifischen Zellularautomatencodes.
 *
 * Alle folgenden Funktionen dienen der interaktiven Benutzeroberfläche.
 */

// changeDrawingColor() wird aufgerufen, wenn der Nutzer im HTML-Dokument einen
// anderen Zelltypen zum Zeichnen auswählt.
function changeDrawingColor(cellType)
{
    // Zur besseren Verständlichkeit:
    var oldCellType = Settings.cellToDraw;

    // Speichere die Änderung nur, wenn es wirklich eine Änderung war:
    if (oldCellType != cellType)
        Settings.cellToDraw = cellType;
}

// WireField.parseChar() und .generateChar() werden von den loadFile()- und
// saveFile()-Routinen in der common.js aufgerufen und verarbeiten beide je ein
// Zeichen der Eingabedatei, bzw. eine Zelle der aktuellen Zellkonfiguration:
WireField.prototype.parseChar = function (row, column, char)
{
    switch (char) {
    case '#':  // '#' sind Elektronenköpfe
        // Wir brauchen die Nachbarwerte nicht zu aktualisieren, dies tut
        // loadFile() global über countAllNeighbours(), deswegen übergeben wir
        // keinen Wert für den Parameter 'update'.
        this.createCell(row, column, 3);
        drawCell(row, column, 3);
        break;
    case '+':  // '+' sind Elektronenenden
        this.createCell(row, column, 2);
        drawCell(row, column, 2);
        break;
    case '*':  // '*' sind Drahtzellen
        this.createCell(row, column, 1);
        drawCell(row, column, 1);
        break;
    case ' ':  // Leerzeichen und alle übrigen Zeichen werden als leere Zellen
    default:   // interpretiert.
        break;
    }
}

WireField.prototype.generateChar = function (row, column)
{
    switch (this.data[row][column].state) {
    case 3:
        return '#';  // '#' sind Elektronenköpfe
        break;
    case 2:
        return '+';  // '+' sind Elektronenenden
        break;
    case 1:
        return '*';  // '*' sind Drahtzellen
        break;
    case 0:
    default:
        return ' ';  // Gib für leere Zellen ein Leerzeichen zurück, dann kann
        break;       // die erzeugte Datei in einem Texteditor betrachtet und
                     // bearbeitet werden.
    }
}

// Die WireField.print()-Methode wurde lediglich während der Entwicklung des
// Programms verwendet, um die interne Zellularautomatensimulation unabhängig
// von der grafischen Ausgabe zu testen. Dazu werden die aktuellen Zellzustände
// als ASCII-Grafik (ähnlich der verwendeten Zeichen beim Speichern in
// Textdateien), sowie die aktuellen Nachbarwerte in zwei Rastern nebeneinander
// ausgegeben. WireField.print() wird im fertigen Programm nicht mehr verwendet
// und ist deswegen auch nicht detailliert kommentiert. Ihre Funktion sollte
// nach dem bisherigen Code dennoch verständlich sein.
WireField.prototype.print = function()
{
    var printstring = '';

    for (var row in this.data) {
        var fieldstring = '';
        var neighbourstring = '';

        for (var column in this.data[row]) {
            switch (this.data[row][column].state) {
            case 3:  // 'head'
                fieldstring += '# ';
                break;
            case 2:  // 'tail'
                fieldstring += '+ ';
                break;
            case 1:  // 'wire'
                fieldstring += '* ';
                break;
            case 0:
            default:
                fieldstring += '  ';
                break;
            }

            neighbourstring += this.data[row][column].neighbours + ' ';
        }
        printstring += fieldstring + '| ' + neighbourstring + '\n ';
    }

    console.log(printstring);
    console.log(this.population + ' Electron heads.')
}
